import contextlib
import glob
import logging
import os
import re
import shutil
import time
import urllib.request

import debian.changelog
import debian.deb822
import debian.debian_support

from mini_buildd import call, changes, config, daemon, files, util

LOG = logging.getLogger(__name__)


class Changelog(debian.changelog.Changelog):
    r"""
    Changelog class with some extra functions

    >>> cl = Changelog(util.fopen("test-data/changelog"), max_blocks=100)
    >>> cl.find_first_not("mini-buildd@buildd.intra")
    ('Stephan Sürken <absurd@debian.org>', '1.0.0-2')

    >>> cl = Changelog(util.fopen("test-data/changelog.ported"), max_blocks=100)
    >>> cl.find_first_not("mini-buildd@buildd.intra")
    ('Stephan Sürken <absurd@debian.org>', '1.0.0-2')

    >>> cl = Changelog(util.fopen("test-data/changelog.oneblock"), max_blocks=100)
    >>> cl.find_first_not("mini-buildd@buildd.intra")
    ('Stephan Sürken <absurd@debian.org>', '1.0.1-1~')

    >>> cl = Changelog(util.fopen("test-data/changelog.oneblock.ported"), max_blocks=100)
    >>> cl.find_first_not("mini-buildd@buildd.intra")
    ('Mini Buildd <mini-buildd@buildd.intra>', '1.0.1-1~')
    """

    def find_first_not(self, author):
        """Find (author,version+1) of the first changelog block not by given author"""
        result = (None, None)
        for index, block in enumerate(self._blocks):
            result = (block.author,
                      f"{self._blocks[index + 1].version}" if len(self._blocks) > (index + 1) else f"{block.version}~")
            if author not in block.author:
                break
        return result


class DebianVersion(debian.debian_support.Version):
    @classmethod
    def _sub_rightmost(cls, pattern, repl, string):
        last_match = None
        for last_match in re.finditer(pattern, string):
            pass
        return string[:last_match.start()] + repl + string[last_match.end():] if last_match else string + repl

    @classmethod
    def _get_rightmost(cls, pattern, string):
        last_match = None
        for last_match in re.finditer(pattern, string):
            pass
        return string[last_match.start():last_match.end()] if last_match else ""

    @classmethod
    def stamp(cls):
        # 20121218151309
        return time.strftime("%Y%m%d%H%M%S", time.gmtime())

    @classmethod
    def stamp_regex(cls, stamp=None):
        return fr"[0-9]{{{len(stamp if stamp else cls.stamp())}}}"

    def gen_internal_rebuild(self):
        r"""
        Generate an 'internal rebuild' version

        If the version is not already a rebuild version, just
        append the rebuild appendix, otherwise replace the old
        one. For example::

          1.2.3 -> 1.2.3+rebuilt20130215100453
          1.2.3+rebuilt20130215100453 -> 1.2.3+rebuilt20130217120517

        Code samples:

        >>> regex = rf"^1\.2\.3\+rebuilt{DebianVersion.stamp_regex()}$"
        >>> bool(re.match(regex, DebianVersion("1.2.3").gen_internal_rebuild()))
        True
        >>> bool(re.match(regex, DebianVersion("1.2.3+rebuilt20130215100453").gen_internal_rebuild()))
        True
        """
        stamp = self.stamp()
        return self._sub_rightmost(r"\+rebuilt" + self.stamp_regex(stamp),
                                   "+rebuilt" + stamp,
                                   self.full_version)

    def gen_external_port(self, default_version):
        """
        Generate an 'external port' version

        This currently just appends the given default version
        appendix. For example:

        1.2.3 -> 1.2.3~test60+1
        """
        return f"{self.full_version}{default_version}"

    def gen_internal_port(self, from_mandatory_version_regex, to_default_version):
        r"""
        Generate an 'internal port' version

        Tests for the (recommended) Default layout:

        >>> sid_regex = r"~testSID\+[1-9]"
        >>> sid_default = "~testSID+1"
        >>> sid_exp_regex = r"~testSID\+0"
        >>> sid_exp_default = "~testSID+0"
        >>> wheezy_regex = r"~test70\+[1-9]"
        >>> wheezy_default = "~test70+1"
        >>> wheezy_exp_regex = r"~test70\+0"
        >>> wheezy_exp_default = "~test70+0"
        >>> squeeze_regex = r"~test60\+[1-9]"
        >>> squeeze_default = "~test60+1"
        >>> squeeze_exp_regex = r"~test60\+0"
        >>> squeeze_exp_default = "~test60+0"

        sid->wheezy ports:

        >>> DebianVersion("1.2.3-1~testSID+1").gen_internal_port(sid_regex, wheezy_default)
        '1.2.3-1~test70+1'
        >>> DebianVersion("1.2.3-1~testSID+4").gen_internal_port(sid_regex, wheezy_default)
        '1.2.3-1~test70+4'
        >>> DebianVersion("1.2.3-1~testSID+4fud15").gen_internal_port(sid_regex, wheezy_default)
        '1.2.3-1~test70+4fud15'
        >>> DebianVersion("1.2.3-1~testSID+0").gen_internal_port(sid_exp_regex, wheezy_exp_default)
        '1.2.3-1~test70+0'
        >>> DebianVersion("1.2.3-1~testSID+0exp2").gen_internal_port(sid_exp_regex, wheezy_exp_default)
        '1.2.3-1~test70+0exp2'

        wheezy->squeeze ports:

        >>> DebianVersion("1.2.3-1~test70+1").gen_internal_port(wheezy_regex, squeeze_default)
        '1.2.3-1~test60+1'
        >>> DebianVersion("1.2.3-1~test70+4").gen_internal_port(wheezy_regex, squeeze_default)
        '1.2.3-1~test60+4'
        >>> DebianVersion("1.2.3-1~test70+4fud15").gen_internal_port(wheezy_regex, squeeze_default)
        '1.2.3-1~test60+4fud15'
        >>> DebianVersion("1.2.3-1~test70+0").gen_internal_port(wheezy_exp_regex, squeeze_exp_default)
        '1.2.3-1~test60+0'
        >>> DebianVersion("1.2.3-1~test70+0exp2").gen_internal_port(wheezy_exp_regex, squeeze_exp_default)
        '1.2.3-1~test60+0exp2'

        No version restrictions: just add default version

        >>> DebianVersion("1.2.3-1").gen_internal_port(".*", "~port+1")
        '1.2.3-1~port+1'
        """
        from_apdx = self._get_rightmost(from_mandatory_version_regex, self.full_version)
        from_apdx_plus_revision = self._get_rightmost(r"\+[0-9]", from_apdx)
        if from_apdx and from_apdx_plus_revision:
            actual_to_default_version = self._sub_rightmost(r"\+[0-9]", from_apdx_plus_revision, to_default_version)
        else:
            actual_to_default_version = to_default_version
        return self._sub_rightmost(from_mandatory_version_regex, actual_to_default_version, self.full_version)


class TemplatePackage(util.TmpDir):
    """Copies a package template into a temporary directory (under 'package/')"""

    def __init__(self, template, package_version=None):
        super().__init__(prefix="template-")
        self.template = template
        self.package_dir = os.path.join(self.tmpdir, "package")
        shutil.copytree(os.path.join(config.PACKAGE_TEMPLATES, template), self.package_dir)
        self.package_version = DebianVersion.stamp() if package_version is None else package_version
        self.package_name = None

    @property
    def dsc(self):
        return glob.glob(os.path.join(self.tmpdir, "*.dsc"))[0]

    @classmethod
    def call(cls, *args, **kwargs):
        environment = call.taint_env(
            {
                "DEBEMAIL": daemon.get_model().email_address,
                "DEBFULLNAME": daemon.get_model().mbd_fullname(),
                "GNUPGHOME": daemon.get().gnupg.home,
            })
        return call.Call(*args, **kwargs, env=environment)


class KeyringPackage(TemplatePackage):
    def __init__(self, **kwargs):
        super().__init__("archive-keyring", **kwargs)

        self.key_id = daemon.get().gnupg.get_first_sec_key()
        LOG.debug("KeyringPackage using key: '%s'", self.key_id)

        self.package_name = f"{daemon.get_model().identity}-archive-keyring"

        # Replace %ID%, %MAINT% and %KEY_ID% in all files in package dir.
        for root, _dirs, _files in os.walk(self.package_dir):
            for f in _files:
                old_file = os.path.join(root, f)
                new_file = old_file + ".new"
                with util.fopen(old_file, "r") as of, util.fopen(new_file, "w") as nf:
                    nf.write(util.subst_placeholders(of.read(),
                                                     {"ID": daemon.get_model().identity,
                                                      "KEY_ID": self.key_id,
                                                      "MAINT": f"{daemon.get_model().mbd_fullname()} <{daemon.get_model().email_address}>"}))
                os.rename(new_file, old_file)

        # Export public GnuPG key into the package
        daemon.get().gnupg.export(os.path.join(self.package_dir, self.package_name + ".gpg"), identity=self.key_id)

        # Generate sources.lists
        for r in util.models().Repository.objects.all():
            for d in r.distributions.all():
                for s in r.layout.suiteoption_set.all():
                    for rb in [None] + list(range(s.rollback)):
                        file_base = f"sources.list.d/{d.base_source.codename}/{r.identity}_{s.suite.name}{'' if rb is None else f'-rollback{rb}'}"
                        for type_, appendix in [("deb", ""), ("deb-src", "_src")]:
                            apt_line = r.mbd_get_apt_line(d, s, rollback=rb)
                            file_name = os.path.join(self.package_dir, f"{file_base}{appendix}.list")
                            os.makedirs(os.path.dirname(file_name), exist_ok=True)
                            with util.fopen(file_name, "w") as f:
                                f.write(apt_line.get(type_) + "\n")

        # Generate dput-ng profile
        with util.fopen(os.path.join(self.package_dir, f"{daemon.get_model().mbd_dput_name()}.json"), "w") as f:
            f.write(daemon.get_model().mbd_get_dputng_profile())

        # Generate SSL certificate
        certs_dir = os.path.join(self.package_dir, "certs")
        os.mkdir(certs_dir)
        if util.http_endpoint().is_ssl():
            with util.fopen(os.path.join(self.package_dir, os.path.join(certs_dir, "mini-buildd.crt")), "w") as f:
                f.write(util.http_endpoint().get_certificate())

        # Generate changelog entry
        self.call(["debchange",
                   "--create",
                   "--package", self.package_name,
                   "--newversion", self.package_version,
                   f"Automatic keyring package for archive '{daemon.get_model().identity}'."],
                  cwd=self.package_dir).check()

        self.call(["dpkg-source",
                   "-b",
                   "package"],
                  cwd=self.tmpdir).check()


class TestPackage(TemplatePackage):
    def __init__(self, template, auto_ports=None):
        super().__init__(template)
        self.package_name = template

        self.call(["debchange",
                   "--newversion", self.package_version,
                   f"Version update '{self.package_version}'."],
                  cwd=self.package_dir).check()
        if auto_ports:
            self.call(["debchange",
                       f"MINI_BUILDD_OPTION: auto-ports={','.join(auto_ports)}"],
                      cwd=self.package_dir).check()
        self.call(["dpkg-source", "-b", "package"], cwd=self.tmpdir).check()


_DEFAULT_PORT_OPTIONS = ["lintian-mode=ignore"]


def _port(dsc_url, source, diststr, version, comments=None, options=None, allow_unauthenticated=False):
    def changelog_entries():
        """Merge comments and options into plain changelog_entries"""
        changelog_entries = [] if comments is None else comments
        for o in _DEFAULT_PORT_OPTIONS if options is None else options:
            changelog_entries.append(f"{changes.Upload.Options.KEYWORD}: {o}")
        return changelog_entries

    with util.tmp_dir(prefix="port-") as tmpdir:
        dfn = files.DebianName(source, version)

        # Download DSC via dget.
        dget = ["dget", "--download-only"]
        if allow_unauthenticated:
            dget.append("--allow-unauthenticated")
        try:
            TemplatePackage.call(dget + [dsc_url], cwd=tmpdir).check()
        except Exception as e:
            raise util.HTTPBadRequest("Given URL can't be downloaded (maybe retry allowing unauthenticated)") from e

        # Get SHA1 of original dsc file
        original_dsc_sha1sum = util.Hash(os.path.join(tmpdir, os.path.basename(dsc_url))).sha1()

        # Unpack DSC (note: dget does not support -x to a dedcicated dir).
        dst = "debian_source_tree"
        TemplatePackage.call(["dpkg-source",
                              "-x",
                              os.path.basename(dsc_url),
                              dst],
                             cwd=tmpdir).check()

        dst_path = os.path.join(tmpdir, dst)

        # Get version and author from original changelog; use the first block not
        with util.fopen(os.path.join(dst_path, "debian", "changelog"), "r") as cl:
            original_author, original_version = Changelog(cl, max_blocks=100).find_first_not(daemon.get_model().email_address)
        LOG.debug("Port: Found original version/author: %s/%s", original_version, original_author)

        # Change changelog in DST
        TemplatePackage.call(["debchange",
                              "--newversion", version,
                              "--force-distribution",
                              "--force-bad-version",
                              "--preserve",
                              "--dist", diststr,
                              f"Automated port via mini-buildd (no changes). Original DSC's SHA1: {original_dsc_sha1sum}."],
                             cwd=dst_path).check()

        for entry in changelog_entries():
            TemplatePackage.call(["debchange",
                                  "--append",
                                  entry],
                                 cwd=dst_path).check()

        # Repack DST
        TemplatePackage.call(["dpkg-source", "-b", dst], cwd=tmpdir).check()
        dsc = os.path.join(tmpdir, dfn.dsc())
        daemon.get().gnupg.sign(dsc)

        # Generate Changes file
        changes_file_path = os.path.join(tmpdir, dfn.changes("source"))
        with util.fopen(changes_file_path, "w+") as out:
            # Note: dpkg-genchanges has a home-brewed options parser. It does not allow, for example, "-v 1.2.3", only "-v1.2.3", so we need to use *one* sequence item for that.
            TemplatePackage.call(["dpkg-genchanges",
                                  "-S",
                                  "-sa",
                                  f"-v{original_version}",
                                  f"-D{util.CField('Originally-Changed-By').fullname}={original_author}"],
                                 cwd=dst_path,
                                 stdout=out).check()

        # Sign and upload
        daemon.get().gnupg.sign(changes_file_path)
        changes.Base(changes_file_path).upload(daemon.get_model().mbd_get_ftp_endpoint(), force=True)
        return {"dsc_url": dsc_url, "source": source, "distribution": diststr, "version": version}


def port(source, from_diststr, to_diststr, version=None, options=None):
    # check from_diststr
    from_repository, from_distribution, from_suite = util.models().parse_diststr(from_diststr)
    from_ls = from_repository.mbd_reprepro.ls(source).filter_all(diststr=from_diststr, version=version, raise_if_not_found="Source to port not found")
    if version is None:
        version = from_ls[from_diststr]["version"]

    # check to_diststr
    to_repository, to_distribution, to_suite = util.models().parse_diststr(to_diststr, check_uploadable=True)

    # Ponder version to use
    v = DebianVersion(version)
    if to_diststr == from_diststr:
        to_version = v.gen_internal_rebuild()
    else:
        to_version = v.gen_internal_port(from_repository.mbd_get_mandatory_version_regex(from_distribution, from_suite),
                                         to_repository.mbd_get_default_version(to_distribution, to_suite))

    return _port(config.URIS["repositories"]["static"].url_join(from_repository.mbd_reprepro.pool.dsc_path(source, version)),
                 source,
                 to_diststr,
                 to_version,
                 options=options,
                 allow_unauthenticated=True)


def port_ext(dsc_url, to_diststr, options=None, allow_unauthenticated=False):
    to_repository, to_distribution, to_suite = util.models().parse_diststr(to_diststr, check_uploadable=True)
    try:
        with urllib.request.urlopen(dsc_url) as response:
            dsc = debian.deb822.Dsc(response)
            source = dsc.get("Source")
            if source is None:
                raise util.HTTPBadRequest("Given URL not a DSC")

            v = DebianVersion(dsc["Version"])
            return _port(dsc_url,
                         source,
                         to_diststr,
                         v.gen_external_port(to_repository.mbd_get_default_version(to_distribution, to_suite)),
                         comments=[f"External port from: {dsc_url}"],
                         options=options,
                         allow_unauthenticated=allow_unauthenticated)
    except urllib.error.URLError as e:
        raise util.HTTPBadRequest(f"{e} ({dsc_url})")


def upload_template_package(template_package, diststr):
    """Portext macro. Used for keyring_packages and test_packages"""
    with contextlib.closing(template_package) as package:
        dsc_url = "file://" + package.dsc
        info = f"Port for {diststr}: {os.path.basename(dsc_url)}"
        try:
            return port_ext(dsc_url, diststr, allow_unauthenticated=True)
        except BaseException as e:
            util.log_exception(LOG, f"FAILED: {info}", e)
            return {"error": str(util.e2http(e)), "dsc_url": dsc_url, "source": None, "distribution": None, "version": None}  # mimics result of port functions



import hashlib
import json
import logging
import mimetypes
import os
import pathlib
import re
import uuid

import django
import django.utils

from mini_buildd import api, config, dist, util

LOG = logging.getLogger(__name__)

register = django.template.Library()


@register.filter
def mbd_sorted(iterable):
    """Just do (python) sorting -- useful for models that use custom python ordering"""
    return sorted(iterable)


@register.simple_tag(takes_context=True)
def mbd_menu_active(context, name):
    return "mbd-menu-active" if context["request"].path == config.URIS[name]["view"].uri else ""


@register.simple_tag
def mbd_get(obj):
    """Identity function. Handy to set variables in templates (``{% mbd_get some_value as myvar %}``)"""
    return obj


@register.simple_tag
def mbd_dict_get(dict_, key, default=None):
    """Get value from dict even if the key has a special name (like hyphens in a changes file field), or is in another variable"""
    return dict_.get(key, default)


@register.filter
def mbd_fromtimestamp(stamp):
    return util.Datetime.from_stamp(stamp)


@register.filter
def mbd_parent(path):
    """
    Get parent path (uri), always w/ trailing slash

    >>> mbd_parent("/parent/file")
    '/parent/'
    >>> mbd_parent("/parent/dir/")
    '/parent/'
    >>> mbd_parent("/parent/dir/dir/")
    '/parent/dir/'
    """
    return str(pathlib.Path(path).parent) + "/"


@register.filter
def mbd_basename(path):
    return os.path.basename(path)


@register.simple_tag
def mbd_join(path, *paths):
    return os.path.join(path, *paths)


@register.filter
def mbd_jsonpp(obj):
    """Get pretty print from json-serializable object"""
    return util.json_pretty(obj)


@register.simple_tag
def mbd_token():
    return f"mbd-token-{uuid.uuid4()}"


@register.simple_tag
def mbd_hash(data):
    return f"mbd-hash-{hashlib.sha224(bytes(data, encoding=config.CHAR_ENCODING)).hexdigest()}"


@register.filter
def mbd_codename(diststr):
    """Get codename from diststr. On any error, just return empty string"""
    try:
        return dist.Dist(diststr).codename
    except Exception:
        return ""


@register.filter
def mbd_repository(diststr):
    """Get repository from diststr. On any error, just return empty string"""
    try:
        return dist.Dist(diststr).repository
    except Exception:
        return ""


@register.simple_tag
def mbd_model_stats(model):
    ret = {}
    model_class = getattr(util.models(), model)
    if getattr(model_class, "mbd_is_prepared", None):
        # Status model
        ret["active"] = model_class.objects.filter(status__exact=model_class.STATUS_ACTIVE).count()
        ret["prepared"] = model_class.objects.filter(status__exact=model_class.STATUS_PREPARED).count()
        ret["removed"] = model_class.objects.filter(status__exact=model_class.STATUS_REMOVED).count()
    ret["total"] = model_class.objects.all().count()
    return ret


@register.simple_tag(takes_context=True)
def mbd_absolute_uri(context, location):
    return context.get("request").build_absolute_uri(location)


@register.inclusion_tag("mini_buildd/includes/tags/api.html", takes_context=True)
def mbd_api(context, call, name=None, hide_options=False, verbose=False, popup=True, **kwargs):
    prefix = "value_"
    api_args = {k[len(prefix):]: v for k, v in kwargs.items() if k.startswith(prefix)}
    if isinstance(call, str):
        api_cls = api.CALLS.get(call)
        if api_cls is None:
            raise util.HTTPBadRequest(f"Unknown API call: {call}")
        api_call = api_cls(**api_args)
    else:
        api_call = call
        api_call.set_args(**api_args)

    args = {"mandatory": [], "optional": [], "hidden": []}
    for arg in api_call.args.values():
        args["mandatory" if arg.needs_value() else "hidden" if hide_options else "optional"].append(arg)
    args["optional" if args["optional"] else "hidden"].append(api.Output())

    return {
        "request": context.get("request"),
        "args": args,
        "call": api_call,
        "name": api_call.name if name is None else name,
        "is_authorized": api_call.AUTH.is_authorized(context.get("user")),
        "verbose": verbose,
        "popup": popup,
    }


@register.inclusion_tag("mini_buildd/includes/tags/sbuild_status.html")
def mbd_sbuild_status(bres):
    return {
        "bres": bres,
        "checks": {
            "lintian": bres.get("sbuild_lintian", "none"),
            "piuparts": bres.get("sbuild_piuparts", "none"),
            "autopkgtest": bres.get("sbuild_autopkgtest", "none"),
        }}


class PlainFilter(dict):
    """
    Enrich plain text with HTML hints (TOC)

    .. note:: http://tools.ietf.org/html/rfc5147: This, in theory, could be some way to create a TOC for text/plain (buildlogs and such). However

      * RFC only describes line/char based links, not text matching based (so scanning the text file would be needed still)
      * As it seems, main browsers simply don't support it
        * https://bugzilla.mozilla.org/show_bug.cgi?id=660583
        * https://bugs.chromium.org/p/chromium/issues/detail?id=77024
    """

    #: Allow (non 'text/') mime types
    ALLOWED_MIME_TYPES = ["application/json"]
    #: Allow extensions that have no official mime type
    ALLOWED_EXTENSIONS = ["buildlog", "changes", "dsc", "buildinfo", "log", "upload"]
    #: Allow complete (base) file names
    ALLOWED_FILES = ["InRelease", "Release", "Sources", "Packages", ".sbuildrc"]

    def ok_to_filter(self, file_path):
        mime_type = mimetypes.guess_type(file_path)[0]
        name = os.path.basename(file_path)
        return \
            mime_type is not None and mime_type.startswith("text/") or \
            mime_type in self.ALLOWED_MIME_TYPES or \
            self["extension"] in self.ALLOWED_EXTENSIONS or \
            name in self.ALLOWED_FILES

    #: Regex filters for ``*.buildlog``.
    BUILDLOG_INDEX_REGEXES = [
        (re.compile(r"\| (.*) \|"), 1, "mbd-info"),                                            # sbuild section headers
        (re.compile(r"^MINI_BUILDD: .+"), 0, "mbd-info"),                                      # mini-buildd banners (see files.py)
        (re.compile(r".*make.*\*\*\*.*"), 0, "mbd-error"),                                     # make fatal error
        (re.compile(r"(^W(ARNING)?: .+)|(^W(ARN)?: .+)", re.IGNORECASE), 0, "mbd-warning"),    # warnings (lintian, others)
        (re.compile(r"(^E(RROR)?: .+)|(^E(RR)?: .+)", re.IGNORECASE), 0, "mbd-error"),         # errors (lintian, others)
        (re.compile(r"(^C(RITICAL)?: .+)|(^C(RIT)?: .+)", re.IGNORECASE), 0, "mbd-error"),     # critical (lintian, others)
    ]

    #: Regex filters for ``*.log``.
    LOG_INDEX_REGEXES = [
        (re.compile(r"^.+ W: (.*)", re.IGNORECASE), 0, "mbd-warning"),
        (re.compile(r"^.+ E: (.*)", re.IGNORECASE), 0, "mbd-error"),
        (re.compile(r"^.+ C: (.*)", re.IGNORECASE), 0, "mbd-error"),
    ]

    def regex_filter(self, file_path, regexes):
        with open(file_path, "r", encoding=config.CHAR_ENCODING, errors="ignore") as f:
            logs = []
            for line in f:
                for m in regexes:
                    match = m[0].match(line)
                    if match:
                        token = mbd_token()
                        logs.append(f"<span id='{token}' class='mbd-plain-toc-item {m[2]}'></span>")
                        title = django.utils.html.escape(match.group(m[1]))
                        self["toc"].append({"class": m[2], "token": token, "title": title})
                        break
                logs.append(django.utils.html.escape(line))
            return django.utils.safestring.mark_safe("".join(logs))

    def filter_buildlog(self, file_path):
        return self.regex_filter(file_path, self.BUILDLOG_INDEX_REGEXES)

    def filter_log(self, file_path):
        return self.regex_filter(file_path, self.LOG_INDEX_REGEXES)

    @classmethod
    def filter_json(cls, file_path):
        with util.fopen(file_path) as f:
            return util.json_pretty(json.load(f))

    @classmethod
    def filter_default(cls, file_path):
        with util.fopen(file_path, "r", errors="ignore") as f:
            return f.read()

    def __init__(self, content, file_path, uri, title):
        super().__init__()
        self["content"] = content
        self["file_path"] = file_path
        self["uri"] = uri
        self["title"] = title
        self["toc"] = []  # May be filled by filters (table of contents)

        if file_path is not None and os.path.exists(file_path):
            self["mtime"] = os.path.getmtime(file_path)

            dummy_, extension = os.path.splitext(file_path)
            self["extension"] = extension[1:]

            if self.ok_to_filter(file_path):
                self["content"] = getattr(self, "filter_" + self["extension"], self.filter_default)(file_path)


@register.inclusion_tag("mini_buildd/includes/tags/file.html")
def mbd_file(content=None, file_path=None, uri=None, title=None):
    return PlainFilter(content, file_path, uri, title)


@register.inclusion_tag("mini_buildd/includes/tags/datatable.html")
def mbd_datatable(identity, order=None, columns="null"):
    return {
        "identity": identity,
        "order": order,
        "columns": columns,
    }



# Generated by Django 4.2.9 on 2024-01-07 17:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0019_migrate_chroot_extra_options'),
    ]

    operations = [
        migrations.AddField(
            model_name='suiteoption',
            name='rollback',
            field=models.IntegerField(default=0),
        ),
    ]



# Generated by Django 4.2.9 on 2024-01-29 15:44

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0025_remove_daemon_show_last_builds'),
    ]

    operations = [
        migrations.RenameField(
            model_name='daemon',
            old_name='show_last_packages',
            new_name='event_queue_size',
        ),
        migrations.RenameField(
            model_name='daemon',
            old_name='build_queue_size',
            new_name='max_parallel_builds',
        ),
        migrations.RenameField(
            model_name='source',
            old_name='codeversion_override',
            new_name='codeversion_guessed',
        ),
    ]



# http://www.sphinx-doc.org/en/stable/config.html
#
# pylint: disable=wrong-import-position,invalid-name  # Disable for file: These checks just can't be satisfied here.

import collections
import glob
import os
import pathlib
import re
import subprocess
import sys

# Add local package dir to path
sys.path.insert(0, os.path.abspath("../src"))

from mini_buildd import api, config, util, webapp  # noqa (pep8 E402)

# Pseudo-configure webapp
webapp.pseudo_configure()

# Autogenerate API documentation for package mini_buildd
# Strange to do this here, but it (works and) seems more coherent than extra support in setup.py
subprocess.check_call(["sphinx-apidoc", "--force", "--separate", "--output-dir", "./python/", "../src/mini_buildd/"])

# Autogenerate static HTML man pages for all commands
os.makedirs("_static/man", exist_ok=True)
COMMANDS = []
for manpage in glob.glob("../src/*.[18]"):
    name = os.path.splitext(os.path.basename(manpage))[0]
    html_file_path = os.path.join("_static/man", name + ".html")
    COMMANDS.append(name)
    print(f"Creating HTML manpage: {manpage} -> {html_file_path}")
    with open(html_file_path, "w", encoding="UTF-8") as f:
        # split magic: Remove CGI header produced by man2html
        f.write(subprocess.check_output(["man2html", "-M", "/static/manual/", manpage]).decode("UTF-8").split("\n", 2)[2])

# Autogenerate admonitions rst (hack)
ADMONITIONS = collections.OrderedDict([
    ("error", ["(BUG) Ongoing issues: No fix yet", []]),
    ("attention", ["(TODO) Ongoing issues: Minor or has workaround", []]),
    ("note", ["(NOTE) Something not obvious you should be aware of", []]),
    ("tip", ["(FAQ) Free tip", []]),
])
ADMONITIONS_FILE = "auto_admonitions.rst"


def pysources():
    return [f".{s}" for s in subprocess.check_output(["./pysources"], cwd="..", encoding="UTF-8").splitlines()]


#: list of tuples: [(file_path, file_hr_id)]
for gen in [
        [pathlib.Path(".").rglob("*.rst"), lambda fn: os.path.splitext(fn)[0]],
        # "../src/mini_buildd/<module>.py" -> "mini_buildd.<module>"
        [pysources(), lambda fn: "python/" + os.path.splitext(str(fn).removeprefix("../src/"))[0].replace("/", ".").replace(".__init__", "")],
        [pathlib.Path("../src/mini_buildd/templates/").rglob("*.html"), lambda fn: "/".join(fn.parts[3:])],
]:
    for f_name in gen[0]:
        if str(f_name) == ADMONITIONS_FILE:
            continue

        with open(f_name, encoding="UTF-8") as f:
            for line in f:
                m = re.match(rf".*\.\. ({'|'.join(ADMONITIONS.keys())}):: (.*)", line)
                if m:
                    print(f"Found admonition: {f_name} ({m[1]}): {m[2]}")
                    ADMONITIONS[m[1]][1].append([gen[1](f_name), m[2]])

with open(ADMONITIONS_FILE, "w", encoding="UTF-8") as f:
    f.write(f"""\
Admonitions (FAQ, NOTE, TODO, BUG)
##################################

Automated index of all admonitions (in this manual as well as in
python code documentation). For simplicity, we only use four
admonitions: *{", ".join(ADMONITIONS)}*.

""")
    for typ, admonitions in ADMONITIONS.items():
        f.write(f"""
.. {typ}:: **{admonitions[0]}**

""")
        for admonition in sorted(admonitions[1]):
            f.write(f"  * `{admonition[0]} <{admonition[0]}.html>`_: {admonition[1]}\n")

# Project data
project = "mini-buildd"
project_copyright = "2012-2024, mini-buildd maintainers"
version = util.__version__
release = util.__version__

# Config
extensions = [
    "sphinx.ext.extlinks",
    "sphinx.ext.autodoc",
    "sphinx.ext.graphviz",
    "sphinx.ext.autosectionlabel",
    "sphinx_ext_mini_buildd",
]
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
exclude_patterns = ["_build"]

rst_epilog = ""
for key, uri in config.URIS.items():
    uri = uri.get("view")
    if uri is not None:
        rst_epilog += f".. _{key}: {uri}\n"

# Style
html_theme = "alabaster"
html_logo = "../src/mini_buildd/static/icons/mini-buildd.svg"
html_favicon = html_logo
html_theme_options = {
    "page_width": "95%",
    "body_max_width": "auto",
    "tip_bg": "lightgreen",
    "note_bg": "lightyellow",
    "globaltoc_maxdepth": 3,
}
html_static_path = ["_static"]
html_sidebars = {
    "**": [
        "globaltoc.html",
        "relations.html",
        "searchbox.html",
        "mbd-sidebar-commands.html",
        "mbd-sidebar-apicalls.html",
    ]
}
html_context = {
    "COMMANDS": COMMANDS,
    "APICALLS": api.CALLS,
}

# Extensions
extlinks = {
    "debbug": ("https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=%s", "Debian Bug #%s"),
    "debpkg": ("https://tracker.debian.org/pkg/%s", "Debian package '%s'"),
    "mbdpage": ("/mini_buildd/%s", "mini-buildd's '%s' page"),
    "mbdcommand": ("_static/man/%s.html", "command '%s'"),
    "apicall": ("/mini_buildd/api/?call=%s", "API call '%s'"),
}
autodoc_member_order = "bysource"   # Order members like in source (sphinx default (strangely enough) seems to be alphabetic order)
autosectionlabel_prefix_document = True

suppress_warnings = [
    "config.cache",  # sphinx 7.3.7: Some config elements (like api.CALLS) can't be pickled/cached. Avoid warnings for that.
]



# Generated by Django 4.2.9 on 2024-02-06 13:02

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0028_migrate_source_codeversion_override'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='source',
            options={},
        ),
    ]



# Generated by Django 4.2.14 on 2024-07-24 15:53

from django.db import migrations

from mini_buildd import dist

LINTIAN_EXTRA_OPTIONS_PREVIOUS_DEFAULT = "--info"


def migrate_lintian_options(apps, _schema_editor):
    Distribution = apps.get_model("mini_buildd", "Distribution")
    for d in Distribution.objects.all():
        codename = dist.Codename(d.base_source.codename)
        setup_options = codename.setup_distribution_options()
        if d.lintian_extra_options == LINTIAN_EXTRA_OPTIONS_PREVIOUS_DEFAULT:
            d.lintian_extra_options = setup_options.get("lintian_extra_options", dist.SETUP["defaults"]["distribution"]["lintian_extra_options"])
        d.lintian_warnfail_options = setup_options.get("lintian_warnfail_options", dist.SETUP["defaults"]["distribution"]["lintian_warnfail_options"])
        d.save()


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0032_distribution_lintian_warnfail_options_and_more'),
    ]

    operations = [
        migrations.RunPython(migrate_lintian_options),
    ]



import contextlib
import getpass
import json
import logging
import sys
import urllib.error

import keyring

from mini_buildd import config, events, net, util

LOG = logging.getLogger(__name__)


class Client():
    def input(self, prompt):
        return input(prompt) if self.interactive else None

    def getpass(self):
        return getpass.getpass(f"Password for {self.endpoint.geturl(with_user=True)}: ") if self.interactive else None

    def __init__(self, endpoint, auto_confirm=False, auto_save_passwords=False, interactive=sys.stdout.isatty() and sys.stdin.isatty()):
        self.endpoint = endpoint if isinstance(endpoint, net.ClientEndpoint) else net.ClientEndpoint(endpoint, protocol="http")
        self.keyring = keyring.get_keyring()
        self.interactive = interactive
        self.auto_confirm = auto_confirm
        self.auto_save_passwords = auto_save_passwords

        LOG.debug("Initialized: %s", self)

    def __str__(self):
        return f"API client: {self.endpoint} [keyring backend={self.keyring.__class__.__qualname__}, auto_confirm={self.auto_confirm}, auto_save_passwords={self.auto_save_passwords}, interactive={self.interactive}]"

    def login(self):
        user = self.endpoint.get_user()
        if user:
            password, save_password = None, False
            try:
                password = self.keyring.get_password(self.endpoint.geturl(), user)
                if not password:
                    password, save_password = self.getpass(), True
            except keyring.errors.KeyringError as e:
                LOG.warning("Keyring error: %s", e)
                password = self.getpass()

            # Login
            self.endpoint.login(password)
            LOG.debug("%s: User '%s' logged in", self.endpoint, user)

            # Logged in: Optionally save credentials
            if save_password and (self.auto_save_passwords or self.input(f"Save password for '{user}' to {self.keyring.__class__.__qualname__}: (Y)es, (N)o? ").upper() == "Y"):
                self.keyring.set_password(self.endpoint.geturl(), user, password)
            password = ""  # Maybe safer
            return True
        return False

    def api(self, command, timeout=None):
        """
        Run API call (interactive)
        """
        LOG.debug("Client.api(): %s", command)

        url, confirm_next_call, skip, result = None, False, False, None
        while result is None and not skip:
            try:
                url = command.url(endpoint=self.endpoint, with_confirm=self.auto_confirm or confirm_next_call, with_output="json")
                confirm_next_call = False
                skip = False
                self.login()
                result = json.load(self.endpoint.urlopen(url, timeout=timeout))
            except BaseException as e:
                try:
                    # Try to get rfc7807 results (embedded in non-200 HTTP responses)
                    rfc7807 = util.Rfc7807.from_json(json.loads(e.read().decode(config.CHAR_ENCODING)))  # pylint: disable=no-member
                    error = util.HTTPError(rfc7807.status, rfc7807.detail)
                except BaseException:
                    error = e

                if not self.interactive:
                    raise error from e

                bf, rs = "\033[1m", "\033[0m"  # terminal candy
                action = self.input(
                    f"\n"
                    f"== {self}\n"
                    f"== API URL: {url}\n"
                    f"{bf}** E: {str(error)}{rs}\n"
                    f"<= [{bf}u{rs}]ser change, [{bf}cC{rs}]onfirm next or all calls, [{bf}s{rs}]kip, [{bf}q{rs}]uit, []retry? "
                )
                if action == "u":
                    self.endpoint.set_user(self.input("Change user: "))
                elif action == "c":
                    confirm_next_call = True
                elif action == "C":
                    self.auto_confirm = True
                elif action in ["s"]:
                    skip = True
                elif action in ["q"]:
                    raise error from e

        return result

    def ievents(self, after=None, types=None, distribution=None, source=None, version=None, minimal_version=None, exit_on=None, fail_on=None):
        with contextlib.closing(self.endpoint.http_connect()) as conn:
            def ihttp_events():
                while True:
                    args = {"after": after.isoformat()} if after is not None else {}
                    conn.request("GET", config.URIS["events"]["attach"].join() + "?" + urllib.parse.urlencode(args))
                    response = conn.getresponse()
                    if response.status != 200:
                        raise util.HTTPError(response.status, response.reason)
                    yield events.Event.from_json(json.load(response))

            yield from events.ifilter(ihttp_events, types=types, distribution=distribution, source=source, version=version, minimal_version=minimal_version, exit_on=exit_on, fail_on=fail_on)

    def event(self, **kwargs):
        return next(self.ievents(**kwargs, exit_on=list(events.Type)))



import contextlib
import datetime
import hashlib
import http
import json
import logging
import multiprocessing
import os
import platform
import re
import shutil
import tempfile
import time
import traceback

import dateparser
import debian.debian_support
import django
import django.core.exceptions
import packaging.version
import twisted

from mini_buildd import config, net, version

LOG = logging.getLogger(__name__)


__version__ = version.__version__


class Versions(dict):
    """Main software component's version support (for compat code, informational)"""

    def __init__(self):
        self.version = __version__
        self["python"] = platform.python_version()
        self["django"] = django.get_version()
        self["twisted"] = twisted.version.public()

    def __str__(self):
        return f"mini-buildd {self.version} ({', '.join([f'{c} {v}' for c, v in self.items()])})"

    def has(self, component, version_string):
        return packaging.version.Version(version_string) <= packaging.version.Version(self[component])


VERSIONS = Versions()


def http_endpoint(number=0):
    return net.ServerEndpoint(config.HTTP_ENDPOINTS[number], protocol="http")


def models():
    """
    Import and return ``models`` module (sort-of dependency injection)

    Use this where you need the ``models`` module but can't do a 'proper import' -- as django needs to be
    configured first.
    """
    from mini_buildd import models as models_module  # pylint: disable=import-outside-toplevel
    return models_module


#: For use in fstrings
NEWLINE = "\n"


def fopen(path, mode="r", **kwargs):
    """
    Text file open with our fixed char encoding (UTF-8)

    UTF-8 may become default for open at some point, but not just yet.

    See https://www.python.org/dev/peps/pep-0597/
    """
    return open(path, mode, encoding=config.CHAR_ENCODING, **kwargs)


class Rfc7807:
    def __init__(self, status, detail=None):
        self.status = status if isinstance(status, http.HTTPStatus) else http.HTTPStatus(status)
        self.detail = "No public details (see service log)" if detail is None else detail

    def __str__(self):
        return f"{self.detail} (HTTP {self.status.value} {self.status.description})"

    def to_json(self):
        return {
            "type": "about:blank",  # No special semantics. Status code is HTTP status code, title is HTTP status message
            "status": self.status.value,
            "title": self.status.description,
            "detail": self.detail,
            "instance": "about:blank",
        }

    @classmethod
    def from_json(cls, data):
        return Rfc7807(data["status"], data["detail"])


class HTTPError(Exception):
    """Public (HTTP) exception -- raise this if the exception string is ok for user consumption"""

    def __init__(self, status, detail=None):
        self.rfc7807 = Rfc7807(status, detail)
        self.status = self.rfc7807.status.value
        self.detail = self.rfc7807.detail
        super().__init__(str(self.rfc7807))

    def __str__(self):
        return str(self.rfc7807)


class HTTPOk(HTTPError):
    def __init__(self, detail=None):
        super().__init__(http.HTTPStatus.OK, detail)


class HTTPNotFound(HTTPError):
    def __init__(self, detail=None):
        super().__init__(http.HTTPStatus.NOT_FOUND, detail)


class HTTPBadRequest(HTTPError):
    def __init__(self, detail=None):
        super().__init__(http.HTTPStatus.BAD_REQUEST, detail)


class HTTPUnauthorized(HTTPError):
    def __init__(self, detail=None):
        super().__init__(http.HTTPStatus.UNAUTHORIZED, detail)


class HTTPUnavailable(HTTPError):
    def __init__(self, detail=None):
        super().__init__(http.HTTPStatus.SERVICE_UNAVAILABLE, detail)


class HTTPShutdown(HTTPUnavailable):
    def __init__(self, detail=None):
        super().__init__("Server Shutdown")


class HTTPInternal(HTTPError):
    def __init__(self, detail=None):
        super().__init__(http.HTTPStatus.INTERNAL_SERVER_ERROR, detail)


def raised_info(exception):
    """Try to get filename:lineno from last raise (i.e., to log even in 'production mode'/exc_info=False)"""
    try:
        tb = exception.__traceback__
        while tb.tb_next is not None:
            tb = tb.tb_next
        return f"{os.path.basename(tb.tb_frame.f_code.co_filename)}:{tb.tb_lineno}"
    except:  # noqa (pep8 E722)  # pylint: disable=bare-except
        return ""


def log_exception(log, message, exception, level=logging.WARNING):
    log.log(level, "%s: %s [raised in: %s]", message, exception, raised_info(exception), exc_info=log.isEnabledFor(logging.DEBUG))


def log_stack(log, level=logging.INFO):
    """Log stack (for debugging)"""
    log.log(level, "".join(traceback.format_stack()))


def e2http(exception, status=http.HTTPStatus.INTERNAL_SERVER_ERROR):
    if isinstance(exception, django.core.exceptions.ValidationError):
        return HTTPBadRequest(" ".join(exception.messages))
    return exception if isinstance(exception, HTTPError) else HTTPError(status)


def rrpes(func, *args, **kwargs):
    """Run ``func``. On exception, return public error str"""
    try:
        return func(*args, **kwargs)
    except BaseException as e:
        log_exception(LOG, f"'{func}' failed (error str returned)", e, level=logging.INFO)
        return str(e2http(e))


def check_program(path, deb=None):
    if not os.path.exists(path):
        raise HTTPUnavailable(f"Program not found: {path} (please install Debian package {deb})")


def singularize(s):
    """Singularize some english nouns from plural"""
    if s:
        is_upper = s[-1].isupper()
        for plural, singular in [("ies", "y"), ("es", "e"), ("s", "")]:
            plural = plural.upper() if is_upper else plural
            singular = singular.upper() if is_upper else singular
            if plural == s[-len(plural):]:
                return s[0:-len(plural)] + singular
    return s


def esplit(s, sep):
    """Unlike python ``str.strip(sep)``, this delivers an empty list for the empty string"""
    return [] if s == "" else s.split(sep)


def uniq(iterable):
    """
    Make list unique

    Sometimes, a ``list`` should be unique, but you don't want to use ``set`` (for example, as it does not keep order or
    can't be JSON-serialized).

    There does not seem to be a simple builtin for this. Since ``python 3.7``, ``dict`` is guaranteed to be ordered, so
    we can use this approach.
    """
    return list(dict.fromkeys(iterable).keys())


def mdget(d, keys, default=None):
    """
    Get nested value from multidimensional dict, return default if any 'path part' is missing

    Rather use this instead of code like ``d.get(foo, {}).get(bar, {}).get(mykey, mydefaultvalue)``.

    >>> d = {"foo": {"bar": {"mykey": "myvalue"}}}
    >>> mdget(d, ["foo", "bar", "mykey"], "default_value")
    'myvalue'
    >>> mdget(d, ["fpp", "bar", "mykey"], "default_value")
    'default_value'
    >>> mdget(d, ["foo", "bsr", "mykey"], "default_value")
    'default_value'
    >>> mdget(d, ["foo", "bar", "mykex"], "default_value")
    'default_value'
    """
    _d = d
    for k in keys:
        _d = _d.get(k)
        if _d is None:
            return default
    return _d


class Datetime:
    """Datetime tools -- always use same/comparable (aware && UTC) datetime objects"""

    EPOCH = datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)
    SINCE_OPTIONS = ["1 hour ago", "1 day ago", "1 week ago", "1 month ago", "1 year ago", "epoch"]

    @classmethod
    def now(cls):
        """Shortcut to be used for all internal UTC stamps"""
        return datetime.datetime.now(datetime.timezone.utc)

    @classmethod
    def from_stamp(cls, stamp):
        return datetime.datetime.fromtimestamp(stamp, tz=datetime.timezone.utc)

    @classmethod
    def from_iso(cls, isostr):
        return datetime.datetime.fromisoformat(isostr).replace(tzinfo=datetime.timezone.utc)

    @classmethod
    def from_path(cls, path):
        return cls.from_stamp(os.path.getmtime(path))

    @classmethod
    def timecode(cls, stamp=None):
        """Timecode string from current timestamp"""
        s = cls.now() if stamp is None else stamp
        return s.strftime("%Y%m%d:%H%M%S:%f")

    @classmethod
    def is_naive(cls, date_obj):
        # https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
        return date_obj.tzinfo is None or date_obj.tzinfo.utcoffset(date_obj) is None

    @classmethod
    def check_aware(cls, date_obj):
        if date_obj and cls.is_naive(date_obj):
            raise HTTPBadRequest("Timestamp not time zone aware")
        return date_obj

    @classmethod
    def parse(cls, date_string, default=None):
        """Use keyword 'epoch' for a big-bang timestamp. Otherwise most formats should work (like '2 weeks ago')"""
        if date_string is None:
            return default
        if date_string == "epoch":
            return cls.EPOCH
        dt = dateparser.parse(date_string)
        if dt is None:
            raise HTTPBadRequest(f"Date parse failed on: '{date_string}' ({Datetime.parse.__doc__})")
        if Datetime.is_naive(dt):
            dt = dt.astimezone()
            LOG.debug("Naive timestamp converted: %s => %s", date_string, dt)
        return dt


class Snake():
    """
    Case style conversion to (lowercase) snake

    >>> Snake("CamelCase").from_camel()
    'camel_case'
    >>> Snake("Kebab-Case").from_kebab()
    'kebab_case'
    """

    __CAMEL2SNAKE = re.compile(r"(?<!^)(?=[A-Z])")

    def __init__(self, name):
        self.name = name

    def from_camel(self):
        return self.__CAMEL2SNAKE.sub("_", self.name).lower()

    def from_kebab(self):
        return self.name.replace('-', '_').lower()


class Field():
    """Changes field name handling (custom prefix && (snake) name conversion)"""

    CPREFIX = "X-Mini-Buildd-"

    def __init__(self, field):
        self.name, self.fullname, self.is_cfield = field, field, False
        if field.startswith(self.CPREFIX):
            self.name = field[len(self.CPREFIX):]
            self.is_cfield = True
        self.snake_name, self.snake_fullname = Snake(self.name).from_kebab(), Snake(self.fullname).from_kebab()


class CField(Field):
    def __init__(self, field):
        super().__init__(self.CPREFIX + field)


class StopWatch():
    def __init__(self):
        self.start = Datetime.now()
        self.stop = None

    def __str__(self):
        return f"{self.start.isoformat(timespec='seconds')} -> {self.start.isoformat(timespec='seconds') if self.stop else 'ongoing'} ({round(self.delta().total_seconds())} seconds)"

    def close(self):
        self.stop = Datetime.now()

    def delta(self):
        return (Datetime.now() if self.stop is None else self.stop) - self.start

    def to_json(self):
        return {"start": self.start.isoformat(),
                "stop": self.stop.isoformat() if self.stop else None,
                "delta": self.delta().total_seconds()}


class TmpDir():
    """Temporary dir class (use with ``with contextlib.closing()...``, or as mixin class)"""

    def __init__(self, **kwargs):
        tmp_route = config.ROUTES.get("tmp")
        self.tmpdir = tempfile.mkdtemp(dir=None if tmp_route is None else tmp_route.path.join(), **kwargs)

    def close(self):
        if "keep" in config.DEBUG:
            LOG.warning("DEBUG MODE('keep'): Skipping removal of temporary dir '%s'", self.tmpdir)
        else:
            if VERSIONS.has("python", "3.12"):
                shutil.rmtree(self.tmpdir, onexc=lambda function, path, excinfo: LOG.warning("Could not remove temporary '%s': %s", path, excinfo))  # pylint: disable=unexpected-keyword-arg
            else:
                shutil.rmtree(self.tmpdir, onerror=lambda function, path, excinfo: LOG.warning("Could not remove temporary '%s': %s", path, excinfo[1]))  # pylint: disable=deprecated-argument


@contextlib.contextmanager
def tmp_dir(**kwargs):
    """Temporary dir context manager (use with ``with``)"""
    tmpdir = TmpDir(**kwargs)
    try:
        yield tmpdir.tmpdir
    finally:
        tmpdir.close()


def nop(*_args, **_kwargs):
    pass


def attempt(func, *args, retval_on_failure=None, **kwargs):
    """
    Run function, warn-log exception on error, but continue

    >>> attempt(lambda x: x, "ypsilon")
    'ypsilon'
    >>> attempt(lambda x: x, "ypsilon", retval_on_failure="xylophon", unknown_arg="xanthippe")
    'xylophon'
    >>> attempt(lambda x: x/0, "ypsilon", retval_on_failure="xylophon")
    'xylophon'
    """
    try:
        return func(*args, **kwargs)
    except BaseException as e:
        log_exception(LOG, f"Attempting '{func}' failed (ignoring)", e, level=logging.WARNING)
        return retval_on_failure


def measure(func, *args, **kwargs):
    start = time.process_time()
    result = func(*args, **kwargs)
    LOG.debug("%s: %s seconds", func, time.process_time() - start)
    return result


def strip_epoch(version_str):
    """Strip the epoch from a version string"""
    return version_str.rpartition(":")[2]


def guess_default_dirchroot_backend(overlay, aufs):
    try:
        release = os.uname()[2]
        # linux 3.18-1~exp1 in Debian removed aufs in favor of overlay
        if debian.debian_support.Version(release) < debian.debian_support.Version("3.18"):
            return aufs
    except BaseException:
        pass

    return overlay


def subst_placeholders(template, placeholders):
    """
    Substitute placeholders in string from a dict

    >>> subst_placeholders("Repoversionstring: %IDENTITY%%CODEVERSION%", { "IDENTITY": "test", "CODEVERSION": "60" })
    'Repoversionstring: test60'
    """
    for key, value in list(placeholders.items()):
        template = template.replace(f"%{key}%", value)
    return template


class Hash():
    """
    Shortcut to get hashsums from file

    >>> Hash("test-data/unix.txt").md5()
    'cc3d5ed5fda53dfa81ea6aa951d7e1fe'
    >>> Hash("test-data/unix.txt").sha1()
    '8c84f6f36dd2230d3e9c954fa436e5fda90b1957'
    """

    def __init__(self, path):
        self.path = path

    def get(self, hash_type="md5"):
        """Get any hash from file contents"""
        hash_func = hashlib.new(hash_type)
        with open(self.path, "rb") as f:
            while True:
                data = f.read(128)
                if not data:
                    break
                hash_func.update(data)
        return hash_func.hexdigest()

    def md5(self):
        return self.get(hash_type="md5")

    def sha1(self):
        return self.get(hash_type="sha1")


def get_cpus():
    try:
        return multiprocessing.cpu_count()
    except BaseException:
        return 1


def list_get(list_, index, default=None):
    try:
        return list_[index]
    except IndexError:
        return default


def rmdirs(path):
    """Remove path recursively -- succeed even if it does not exist in the first place"""
    if os.path.exists(path):
        shutil.rmtree(path)
        LOG.info("Directory removed recursively: %s", path)


def json_pretty(json_obj):
    return json.dumps(json_obj, indent=2)



import docutils.nodes
import docutils.parsers.rst


class Example(docutils.parsers.rst.Directive):
    def run(self):
        paragraph_node = docutils.nodes.paragraph(text="Example")
        return [paragraph_node]


def setup(app):
    app.add_directive("mbd-example", Example)
    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# Generated by Django 4.2.8 on 2024-01-03 16:42

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0007_remove_auto_reactivate'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='daemon',
            name='hostname',
        ),
    ]



# Generated by Django 4.2.9 on 2024-01-07 16:14

from django.db import migrations

from . import mbd_get_extra_option


def copy_extra_options(apps, _schema_editor):
    Daemon = apps.get_model("mini_buildd", "Daemon")
    for daemon in Daemon.objects.all():
        daemon.custom_archive_origin = mbd_get_extra_option(daemon, "X-Archive-Origin", "")
        daemon.save()


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0016_add_daemon_extra_options'),
    ]

    operations = [
        migrations.RunPython(copy_extra_options),
    ]



import logging
from contextlib import closing

from mini_buildd import changes, daemon, events, package, threads, util

LOG = logging.getLogger(__name__)


class Package():
    def __init__(self, upload):
        self.upload = upload

        self.buildrequests, self.buildresults = {}, {}

        # Get/check repository, distribution and suite for changes
        self.repository, self.distribution, self.suite = util.models().parse_dist(self.upload.dist, check_uploadable=True)

        # Authenticate
        if self.repository.allow_unauthenticated_uploads:
            LOG.warning("Unauthenticated uploads allowed. Using '%s' unchecked", self.upload.file_name())
        else:
            with closing(daemon.UploadersKeyring(self.repository.identity)) as gpg:
                gpg.verify(self.upload.file_path)

        # Repository package prechecks
        self.repository.mbd_package_precheck(self.distribution, self.suite, self.upload["source"], self.upload["version"])

    def __str__(self):
        return self.upload.key

    def upload_buildrequests(self):
        self.buildrequests = self.upload.request_builds(self.repository, self.distribution, self.suite)
        daemon.get().events.log(events.Type.PACKAGING, self.upload, extra={"buildrequests": {breq.get("architecture"): breq.to_event_json() for breq in self.buildrequests.values()}})

    def add_buildresult(self, bres):
        with closing(daemon.RemotesKeyring()) as gpg:
            gpg.verify(bres.file_path)
        self.buildresults[bres["Architecture"]] = bres

    def auto_ports(self):
        """Try to serve all automatic ports. This will never fail -- failures will be event-logged only"""
        info = {}
        for to_dist_str in set(self.upload.options.get("auto-ports", default=[])):
            try:
                info[to_dist_str] = (True, package.port(self.upload["Source"],
                                                        self.upload.dist.get(),
                                                        to_dist_str,
                                                        self.upload["Version"]))
            except BaseException as e:
                util.log_exception(LOG, "Auto port failed", e)
                info[to_dist_str] = (False, f"{util.e2http(e)}")
        return info

    def finish(self):
        if self.buildrequests.keys() == self.buildresults.keys():
            succeeded = {arch: bres for arch, bres in self.buildresults.items() if bres.success(self.upload, self.distribution, ignore_checks=self.suite.experimental)}
            extra = {}

            status = events.Type.FAILED
            try:
                self.repository.mbd_package_install(self.distribution, self.suite, self.upload, succeeded)
                status = events.Type.INSTALLED
                extra.setdefault("auto_ports", self.auto_ports())
            except BaseException as e:
                extra["error"] = util.e2http(e).rfc7807.to_json()
            finally:
                extra["buildresults"] = {arch: bres.to_event_json() for arch, bres in self.buildresults.items()}
                self.upload.move_to(self.upload.get_events_path())
                for c in list(self.buildresults.values()) + list(self.buildrequests.values()):
                    c.move_to(c.get_events_path())
                daemon.get().events.log(status, self.upload, extra=extra)
            return True
        return False


class Packager(threads.EventThread):
    def __init__(self):
        super().__init__()
        self.packaging = {}

    def run_event(self, event):
        try:
            with self.lock:
                if isinstance(event, changes.Buildresult):
                    if event.key not in self.packaging:
                        raise util.HTTPBadRequest(f"Stray buildresult (not packaging here): {event.bkey}")

                    pkg = self.packaging[event.key]
                    pkg.add_buildresult(event)
                    if pkg.finish():
                        del self.packaging[pkg.upload.key]

                elif isinstance(event, changes.Upload):
                    pkg = Package(event)
                    if pkg.upload.key in self.packaging:
                        raise util.HTTPInternal(f"Stray upload (already packaging): {pkg.upload.key}")  # Incoming handler should always prevent this in the 1st place
                    pkg.upload_buildrequests()
                    self.packaging[pkg.upload.key] = pkg
                else:
                    raise util.HTTPInternal(f"Package Queue: Only ``Upload`` or ``Buildresult`` instances allowed: {event}")
        except Exception as e:
            daemon.get().events.log(events.Type.REJECTED, event, e)



# Generated by Django 4.2.9 on 2024-02-01 14:38

from django.db import migrations


# Keep old override behaviour for existing instances.
# Previously, codeversion in essence behaved like codeversion_override
def migrate_source_codeversion_override(apps, _schema_editor):
    Source = apps.get_model("mini_buildd", "Source")
    for source in Source.objects.all():
        source.codeversion_override = source.codeversion
        source.save()


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0027_source_codeversion_override'),
    ]

    operations = [
        migrations.RunPython(migrate_source_codeversion_override),
    ]



# Generated by Django 4.2.9 on 2024-01-07 14:31

from django.db import migrations

from . import mbd_get_extra_option

SBUILD2INT = {
    "DISABLED": 0,
    "0": 0,
    "IGNORE": 1,
    "1": 1,
    "ERRFAIL": 2,
    "2": 2,
    "WARNFAIL": 3,
    "3": 3,
}


def copy_extra_options(apps, _schema_editor):
    Distribution = apps.get_model("mini_buildd", "Distribution")
    for distribution in Distribution.objects.all():
        distribution.add_depends = mbd_get_extra_option(distribution, "Add-Depends", "")
        distribution.autopkgtest_mode = SBUILD2INT.get(mbd_get_extra_option(distribution, "Autopkgtest-Mode", "0"), 0)
        distribution.deb_build_options = mbd_get_extra_option(distribution, "Deb-Build-Options", "")
        distribution.deb_build_profiles = mbd_get_extra_option(distribution, "Deb-Build-Profiles", "")
        distribution.internal_apt_priority = int(mbd_get_extra_option(distribution, "Internal-APT-Priority", "1"))
        distribution.sbuild_config_blocks = mbd_get_extra_option(distribution, "Sbuild-Config-Blocks", "")
        distribution.sbuild_setup_blocks = mbd_get_extra_option(distribution, "Sbuild-Setup-Blocks", "")
        distribution.save()


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0014_add_distribution_extra_options'),
    ]

    operations = [
        migrations.RunPython(copy_extra_options),
    ]



# Generated by Django 4.2.9 on 2024-01-06 17:37

from django.db import migrations, models
import mini_buildd.models


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0010_migrate_source_extra_options'),
    ]

    operations = [
        migrations.AddField(
            model_name='layout',
            name='meta_distributions',
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='layout',
            name='extra_options',
            field=mini_buildd.models.UnixTextField(blank=True),
        ),
    ]



import logging
import os
import shlex
import subprocess
import time

from mini_buildd import config, util

LOG = logging.getLogger(__name__)


def taint_env(taint):
    env = os.environ.copy()
    for name in taint:
        env[name] = taint[name]
    return env


class Call():
    """
    Wrapper around python subprocess

    When supplying ``stdout`` or ``stderr``, provide raw and
    'seekable' file-like object; i.e., use "w+" and standard
    python ``open`` like::

      mystdout = open(myoutputfile, "w+")

    >>> Call(["echo", "-n", "hallo"]).check().stdout
    'hallo'

    >>> Call(["ls", "__no_such_file__"]).check()  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    util.HTTPBadRequest: Internal call failed (HTTP 400 Bad request syntax or unsupported method)

    >>> Call(["printf stdin; printf stderr >&2"], stderr=subprocess.STDOUT, shell=True).stdout
    'stdinstderr'
    """

    @classmethod
    def _call2shell(cls, call):
        """
        Convert call to human-readable shell line

        Converts an argument sequence ("call") to a
        command line more human-readable and "likely-suitable"
        for cut and paste to a shell.
        """
        return shlex.join(call)

    def __init__(self, call, run_as_root=False, **kwargs):
        self.call = ["sudo", "-n"] + call if run_as_root else call
        self.kwargs = kwargs.copy()

        # Generate stdout and stderr args if not given explicitly, and remember streams given explicitly
        self._given_stream = {}
        for stream in ["stdout", "stderr"]:
            if stream not in self.kwargs:
                self.kwargs[stream] = subprocess.PIPE
            elif not isinstance(self.kwargs[stream], int):
                self._given_stream[stream] = self.kwargs[stream]

        self.result = subprocess.run(self.call, **self.kwargs, check=False)

        # Convenience 'label' for log output
        self.label = f"{'#' if run_as_root else '?'} {call[0]}.."

        LOG.debug("Called with returncode %s: %s", self.result.returncode, self._call2shell(self.call))

    @classmethod
    def _bos2str(cls, value, errors="replace"):
        """Return str(), regardless if value is of type bytes or str"""
        return value if isinstance(value, str) else value.decode(encoding=config.CHAR_ENCODING, errors=errors)

    def _stdx(self, value, key):
        """Stdin or stdout value as str"""
        if value:
            return self._bos2str(value)
        if key in self._given_stream:
            self._given_stream[key].seek(0)
            return self._bos2str(self._given_stream[key].read())
        return ""

    @property
    def stdout(self):
        """Stdout value (empty string if none)"""
        return self._stdx(self.result.stdout, "stdout")

    @property
    def stderr(self):
        """Stderr value (empty string if none)"""
        return self._stdx(self.result.stderr, "stderr")

    def log(self, level=None):
        """
        Log calls output to mini-buildd's logging for debugging

        Per default, this logs to level ``error`` on failure, ``debug`` on success.
        """
        level = level if level is not None else logging.DEBUG if self.result.returncode == 0 else logging.ERROR
        for prefix, output in [("stdout", self.stdout), ("stderr", self.stderr)]:
            for line in output.splitlines():
                LOG.log(level, "%s (%s): %s", self.label, prefix, line)
        return self

    def success(self):
        return self.result.returncode == 0

    def check(self, public_message="Internal call failed"):
        """Raise on unsuccessful (returncode != 0) call"""
        if not self.success():
            self.log()
            raise util.HTTPBadRequest(public_message)
        return self


def call_sequence(calls, run_as_root=False, rollback_only=False, **kwargs):
    """
    Run sequences of calls with rollback support

    >>> call_sequence([(["echo", "-n", "cmd0"], ["echo", "-n", "rollback cmd0"])])
    >>> call_sequence([(["echo", "cmd0"], ["echo", "rollback cmd0"])], rollback_only=True)
    """
    def rollback(pos):
        for i in range(pos, -1, -1):
            if calls[i][1]:
                Call(calls[i][1], run_as_root=run_as_root, **kwargs).log()
            else:
                LOG.debug("Skipping empty rollback call sequent %s", i)

    if rollback_only:
        rollback(len(calls) - 1)
    else:
        i = 0
        try:
            for c in calls:
                if c[0]:
                    Call(c[0], run_as_root=run_as_root, **kwargs).check()
                else:
                    LOG.debug("Skipping empty call sequent %s", i)
                i += 1
        except BaseException:
            LOG.error("Sequence failed at: %s (rolling back)", i)
            rollback(i)
            raise


def call_with_retry(call, retry_max_tries=5, retry_sleep=1, retry_failed_cleanup=None, **kwargs):
    """
    Run call repeatedly until it succeeds (retval 0)

    In case retry_max_tries is reached, the error from the last try is raised.

    >>> call_with_retry(["/bin/true"])
    >>> call_with_retry(["/bin/false"])  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
      ...
    util.HTTPBadRequest: Internal call failed (HTTP 400 Bad request syntax or unsupported method)
    """
    for t in range(retry_max_tries):
        try:
            Call(call, **kwargs).check()
            return
        except BaseException as e:
            if t >= retry_max_tries - 1:
                raise
            LOG.warning("Retrying call in %s seconds [retry #{t}]: %s", retry_sleep, e)
            if retry_failed_cleanup is not None:
                retry_failed_cleanup()
            time.sleep(retry_sleep)



import collections
import datetime
import enum
import functools
import glob
import json
import logging
import operator
import os.path
import queue
import threading
import weakref

import debian.debian_support

from mini_buildd import call, config, daemon, files, util

LOG = logging.getLogger(__name__)


class Type(enum.Enum):
    REJECTED = enum.auto()

    BUILDING = enum.auto()
    BUILT = enum.auto()

    PACKAGING = enum.auto()
    INSTALLED = enum.auto()
    FAILED = enum.auto()

    MIGRATED = enum.auto()
    REMOVED = enum.auto()


#: Human-readable semantics of all event types.
DESC = {
    Type.REJECTED: "User upload was rejected",

    Type.BUILDING: "Building on one of our chroots has started",
    Type.BUILT: "Building on one of our chroots has finished (does not imply a successful build result)",

    Type.PACKAGING: "Packaging to one of our repositories has started",
    Type.INSTALLED: "Packaging to one of our repositories was successful",
    Type.FAILED: "Packaging to one of our repositories has failed",

    Type.MIGRATED: "A package in one of our repositories has been migrated",
    Type.REMOVED: "A package in one of our repositories has been removed",
}


class Event():
    #: Set of "building" event types (package not (necessarily) in instance's repository)
    BUILDING = {Type.BUILDING, Type.BUILT}

    #: Set of "packaging" event types (package in instances's repository)
    PACKAGING = set(Type) - BUILDING

    def __init__(self, type_, distribution, source, version, extra):
        self.type = type_
        self.distribution = distribution
        self.source = source
        self.version = version
        self.timestamp = util.Datetime.now()
        self.extra = extra

    def strerror(self):
        """Public human-readable one-liner string from ``extra.error``"""
        error = self.extra.get("error")  # Rfc7807 dict
        return "" if error is None else error.get("detail", "No details")

    def desc(self):
        return DESC[self.type]

    def __str__(self):
        return ": ".join(filter(None, [f"{self.type.name}({self.distribution}): {self.source}-{self.version} ({self.extra.get('architecture')})", self.strerror()]))

    def match(self, types=None, distribution=None, source=None, version=None, minimal_version=None):
        def match(match, value):
            return match is None or match == value

        return \
            (types is None or self.type in types) and \
            match(distribution, self.distribution) and \
            match(source, self.source) and \
            match(version, self.version) and \
            (minimal_version is None or debian.debian_support.Version(self.version) >= debian.debian_support.Version(minimal_version))

    def to_json(self):
        return {
            "type": self.type.name,
            "distribution": self.distribution,
            "source": self.source,
            "version": self.version,
            "timestamp": self.timestamp.isoformat(),
            "extra": self.extra,
        }

    @classmethod
    def from_json(cls, data):
        event = Event(Type[data["type"]],
                      data["distribution"],
                      data["source"],
                      data["version"],
                      data["extra"])
        event.timestamp = datetime.datetime.fromisoformat(data["timestamp"])
        return event

    @classmethod
    def from_changes(cls, typ, changes, exception=None, extra=None):
        xtra = {}
        xtra.update(changes.to_event_json())
        if exception is not None:
            xtra["error"] = util.e2http(exception).rfc7807.to_json()
        if extra is not None:
            xtra.update(extra)

        return Event(typ,
                     distribution=changes.get("distribution"),
                     source=changes.get("source"),
                     version=changes.get("version"),
                     extra=xtra)

    def save_as(self, file_path):
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with util.fopen(file_path, "w") as f:
            json.dump(self.to_json(), f)

        hook_pattern = "[!_.]*[!~]"
        for hook in sorted(glob.glob(config.ROUTES["event_hooks"].path.join(hook_pattern))):
            if os.path.isfile(hook) and os.access(hook, os.X_OK):
                try:
                    call.Call([hook, file_path]).check()
                except BaseException as e:
                    LOG.warning("Custom Event Hook failed: %s: %s", hook, e)

    def json_file_name(self):
        return files.DebianName(self.source, self.version).json(self.type.name, self.extra["architecture"] if self.type in [Type.BUILT, Type.BUILDING] else None)

    @classmethod
    @functools.lru_cache(maxsize=512)
    def load(cls, file_path):
        with util.fopen(file_path) as f:
            return Event.from_json(json.load(f))

    def ongoing(self):
        """Get ongoing status (bool) -- convenience for template access"""
        try:
            return self.extra["bkey"] in daemon.get().events.ongoing
        except Exception:
            return False

    def isummary(self):
        """Produce most noteworthy/usable textual information from event (for example, for an email body)"""
        if self.type in [Type.REJECTED]:
            pass
        elif self.type in [Type.BUILDING]:
            yield "Live buildlog", self.extra.get("buildlog_building")
        elif self.type in [Type.BUILT]:
            yield "Buildlog", self.extra.get("buildlog_installed")
        elif self.type in [Type.PACKAGING, Type.INSTALLED, Type.FAILED]:
            changes = self.extra.get("changes")
            if changes is not None:
                yield "Changes", config.URIS["events"]["dir"].url_join(changes).replace(" ", "%20")

            buildresults = self.extra.get("buildresults")
            if buildresults is not None:
                for arch, bres in self.extra["buildresults"].items():
                    yield f"Buildlog({arch})", bres["buildlog_installed"]

            buildrequests = self.extra.get("buildrequests")
            if buildrequests is not None:
                for arch, bres in self.extra["buildrequests"].items():
                    yield f"Live buildlog({arch})", bres["buildlog_building"]

            auto_ports = self.extra.get("auto_ports")
            if auto_ports is not None:
                for d, port in auto_ports.items():
                    yield f"Autoport({d})", port[1]["dsc_url"] if port[0] else port[1]
        elif self.type in [Type.MIGRATED]:
            pass
        elif self.type in [Type.REMOVED]:
            pass

    def summary(self):
        s = dict(self.isummary())
        ml = max((len(name) for name in s.keys()), default=0)  # "max length" formatting helper
        return "\n".join(f"{name:<{ml}}: {info}" for name, info in self.isummary())


class Queue(collections.deque):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._clients = {}
        self._lock = threading.Lock()
        self.ongoing = collections.OrderedDict()

    def log(self, typ, changes, exception=None, extra=None):
        if exception is None and changes.cget("Internal-Error"):
            exception = util.HTTPInternal(changes.cget("Internal-Error"))

        if exception is not None:
            util.log_exception(LOG, "Event exception", exception)

        event = Event.from_changes(typ, changes, exception, extra)
        file_path = changes.get_events_path(event.json_file_name())
        event.save_as(file_path)
        LOG.debug("Logging event: %s (%s)", event, file_path)
        with self._lock:
            self.appendleft(event)
            util.attempt(daemon.get_model().mbd_notify_event, event)

            for _key, (weakref_obj, q) in self._clients.items():
                if weakref_obj() is not None:
                    LOG.debug("Queuing event for object '%s'", weakref_obj)
                    q.put(event)

            # Handle "ongoing" dict
            if event.type in [Type.PACKAGING, Type.BUILDING]:
                LOG.debug("Adding event to 'ongoing': %s=%s", changes.bkey, event)
                self.ongoing[changes.bkey] = event
            elif event.type in [Type.BUILT, Type.INSTALLED, Type.FAILED]:
                LOG.debug("Deleting event from 'ongoing': %s from %s", changes.bkey, event)
                self.ongoing.pop(changes.bkey, None)   # None: Should rather not raise exception here if key is not there for some reason

    def attach(self, obj, after=None):
        with self._lock:
            # Garbage collection
            for key in [key for key, (weakref_obj, _q) in self._clients.items() if weakref_obj() is None]:
                del self._clients[key]
                LOG.debug("Events queue client detached (%s/%s): %s", len(self._clients), config.MAX_EVENTS_QUEUE_CLIENTS, obj)

            if id(obj) not in self._clients:
                if len(self._clients) >= config.MAX_EVENTS_QUEUE_CLIENTS:
                    raise util.HTTPUnavailable(f"Max client limit ({config.MAX_EVENTS_QUEUE_CLIENTS}) exceeded")
                self._clients[id(obj)] = (weakref.ref(obj), queue.Queue())
                LOG.debug("Events queue client attached (%s/%s): %s", len(self._clients), config.MAX_EVENTS_QUEUE_CLIENTS, obj)

                # Optionally put past events
                if after is not None:
                    for event in reversed(self):
                        if event.timestamp > after:
                            self._clients[id(obj)][1].put(event)

            return self._clients[id(obj)][1]

    def shutdown(self):
        """Hint shutdown to all client (queues). Essentially makes blocking get() in httpd.py continue so httpd can shutdown"""
        with self._lock:
            for client in self._clients.values():
                try:
                    client[1].put_nowait(config.SHUTDOWN)
                except Exception as e:
                    LOG.warning("Events queue: Ignoring failure to hint shutdown to client %s: %s", client[0], e)

    def to_json(self):
        return [event.to_json() for event in self]

    @classmethod
    def from_json(cls, events, maxlen):
        return Queue([Event.from_json(event) for event in events], maxlen=maxlen)

    @classmethod
    def load(cls, maxlen):
        return Queue(load(maxlen=maxlen), maxlen=maxlen)


def get(events):
    event = events.get()
    if event is config.SHUTDOWN:
        raise util.HTTPShutdown()
    return event


class Attach():
    def __init__(self, events, *args, **kwargs):
        self.events = events.attach(self, *args, **kwargs)

    def get(self):
        return get(self.events)


def ifilter(ievents, types=None, distribution=None, source=None, version=None, minimal_version=None, exit_on=None, fail_on=None):
    for event in ievents():
        if event.match(types=types, distribution=distribution, source=source, version=version, minimal_version=minimal_version):
            yield event
            if fail_on is not None and event.type in fail_on:
                raise util.HTTPBadRequest(f"Failed on event: {event}")
            if exit_on is not None and event.type in exit_on:
                break


def load(path="", after=None, before=None, maxlen=None):
    """Load events (as normal list) from events path, ordered as issued using event's timestamp"""
    return sorted([Event.load(entry.path) for entry in config.ROUTES["events"].path.ifiles(path, ".json", after=after, before=before)], key=operator.attrgetter("timestamp"), reverse=True)[0:maxlen]



# Generated by Django 4.2.9 on 2024-01-07 16:09

from django.db import migrations, models
import mini_buildd.config


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0015_migrate_distribution_extra_options'),
    ]

    operations = [
        migrations.AddField(
            model_name='daemon',
            name='custom_archive_origin',
            field=models.CharField(blank=True, default='', max_length=255),
        ),
        migrations.AlterField(
            model_name='daemon',
            name='identity',
            field=models.CharField(default=mini_buildd.config.default_identity, max_length=50),
        ),
    ]



r"""
.. attention:: **wheezy or older builds on amd64**: Re-enable linux' vsyscall

  In Debian kernel packages since ``4.8.4-1~exp1``::

    [ Ben Hutchings ]
    * [amd64] Enable LEGACY_VSYSCALL_NONE instead of LEGACY_VSYSCALL_EMULATE.
      This breaks (e)glibc 2.13 and earlier, and can be reverted using the kernel
      parameter: vsyscall=emulate

  I.e.: When running the Debian standard kernel and your mini-buildd instance needs to support ``wheezy`` or earlier, you
  need to re-enable this (in ``/etc/default/grub``).

  You may use ``local-vsyscall-emulate.cfg`` in ``examples`` (a symlink should do) for this purpose.

  On any running kernel, this is a poor man's check to see if vsyscall is still enabled on your system:

  .. code:: bash

    grep "\[vsyscall\]" /proc/self/maps

  .. seealso:: :debbug:`847154`, linux package's ``NEWS`` file.

.. note:: **schroot**: Verbosely logs commands ("Running command ...") explicitly via syslog (journalctl)

  schroot seems to verbosely log any executed root-user command via syslog (user/notice). This will
  clutter journalctl's (standard) log output via unit::

    journalctl --unit mini-buildd.service --follow

  As a workaround, specifying the (syslog compatibility) identifier like so helps to unclutter::

    journalctl --identifier mini-buildd --follow
"""

import logging

from mini_buildd import call

LOG = logging.getLogger(__name__)

COMMAND = "/usr/bin/schroot"


class Session():
    def __init__(self, name, namespace="chroot"):
        self.chroot = f"{namespace}:{name}"
        self.session = "session:" + call.Call([COMMAND,
                                               "--begin-session",
                                               "--chroot", self.chroot]).check().stdout.strip()
        LOG.debug("%s: Schroot session started ({self.chroot})", self)

    def close(self):
        """
        Close session (including retries on failure)

        .. attention:: **schroot**: 'target is busy' on session close (stale schroot sessions)

           Occasionally, a schroot session can't be closed
           properly, leaving stale sessions around. Presumably,
           external programs (like 'desktop mount scanners') can cause
           this.

           This internal close does try hard to avoid this -- however,
           if disaster strikes anyway, ``mini-buildd-cruft``
           may help to remove these stale sessions manually (i.e., as
           ``mini-buildd`` user from the shell).
        """
        call.call_with_retry([COMMAND,
                              "--end-session",
                              "--chroot", self.session,
                              ],
                             retry_max_tries=10,
                             retry_sleep=1)
        LOG.debug("%s: Schroot session closed (%s)", self, self.chroot)

    def call(self, _call, user="root"):
        return call.Call([COMMAND,
                          "--run-session",
                          "--chroot", self.session,
                          "--user", user] + _call)

    def run(self, _call, user="root"):
        return self.call(_call, user=user).check().stdout

    def info(self):
        return call.Call([COMMAND,
                          "--info",
                          "--chroot", self.session]).check().stdout

    def __str__(self):
        return f"{self.session}"

    def update_file(self, file_path, content):
        """Write content to file"""
        self.run(["--", "/bin/sh", "-c", f"echo '{content}' >{file_path}"])
        LOG.debug("%s: Updated file '%s': %s", self, file_path, content)

    def set_debconf(self, key, value):
        """Set arbitrary debconf value"""
        self.run(["--", "/bin/sh", "-c", f"export DEBIAN_FRONTEND=noninteractive && echo 'set {key} {value}' | debconf-communicate"])
        LOG.debug("%s: Debconf value set: %s=%s", self, key, value)
